#include "cadical.hpp"

#include <cstdlib>
#include <cstring>

namespace CaDiCaL {

struct Wrapper : Learner, Terminator {
  bool lastSat = 0;
  bool lastUnsat = 0;
  int nCalls = 0;
  int nSat = 0;
  int nUnsat = 0;
  int nVal = 0;
  int nFailed = 0;
  int readModel = 0;
  int readCore = 0;
  int cumVal = 0;
  int cumFailed = 0;
  float cumPVal = 0;
  float cumPFailed = 0;

  Solver * solver;
  struct {
    void * state;
    int (*function) (void *);
  } terminator;

  struct {
    void * state;
    int max_length;
    int * begin_clause, * end_clause, * capacity_clause;
    void (*function) (void *, int *);
  } learner;

  long runtime_limit = 0;
  int run_id         = 0;
  std::vector<int> assumptions;

  bool terminate() {
    if (!terminator.function)
      return false;
    return terminator.function (terminator.state);
  }

  bool learning (int size) {
    if (!learner.function)
      return false;
    return size <= learner.max_length;
  }

  void learn (int lit) {
    if (learner.end_clause == learner.capacity_clause) {
      size_t count = learner.end_clause - learner.begin_clause;
      size_t size = count ? 2*count : 1;
      learner.begin_clause = (int*)
        realloc (learner.begin_clause, size * sizeof (int));
      learner.end_clause = learner.begin_clause + count;
      learner.capacity_clause = learner.begin_clause + size;
    }
    *learner.end_clause++ = lit;
    if (lit)
      return;
    learner.function (learner.state, learner.begin_clause);
    learner.end_clause = learner.begin_clause;
  }

  Wrapper () : solver (new Solver ()) {
    memset (&terminator, 0, sizeof terminator);
    memset (&learner, 0, sizeof learner);
  }

  ~Wrapper () {
    terminator.function = 0;
    if (learner.begin_clause) free (learner.begin_clause);
    delete solver; }
};

}

using namespace CaDiCaL;

extern "C" {

#include "ccadical.h"

const char * ccadical_signature (void) {
  return Solver::signature ();
}

CCaDiCaL * ccadical_init (void) {
  return (CCaDiCaL*) new Wrapper ();
}

void ccadical_release (CCaDiCaL * wrapper) {
  delete (Wrapper*) wrapper;
}

void ccadical_set_option (CCaDiCaL * wrapper,
                          const char * name, int val) {
  ((Wrapper*) wrapper)->solver->set (name, val);
}

void ccadical_limit (CCaDiCaL * wrapper,
                     const char * name, int val) {
  ((Wrapper*) wrapper)->solver->limit (name, val);
}

int ccadical_get_option (CCaDiCaL * wrapper, const char * name) {
  return ((Wrapper*) wrapper)->solver->get (name);
}

void ccadical_set_activation (CCaDiCaL * wrapper, int lit){
  ((Wrapper*) wrapper)->solver->set_activation (lit);
}

void ccadical_add (CCaDiCaL * wrapper, int lit) {
  ((Wrapper*) wrapper)->solver->add (lit);
}

void ccadical_assume (CCaDiCaL * wrapper, int lit) {
  ((Wrapper *)wrapper)->assumptions.push_back(lit);
  ((Wrapper*) wrapper)->solver->assume (lit);
}

int ccadical_solve (CCaDiCaL * wrapper) {
  const bool cModel = (((Wrapper *)wrapper)->lastSat) && (((Wrapper *)wrapper)->nVal);
  const bool cCore = (((Wrapper *)wrapper)->lastUnsat) && (((Wrapper *)wrapper)->nFailed);
  ((Wrapper *)wrapper)->readModel += cModel;
  ((Wrapper *)wrapper)->readCore += cCore;
  ((Wrapper *)wrapper)->cumVal += ((Wrapper *)wrapper)->nVal;
  ((Wrapper *)wrapper)->cumFailed += ((Wrapper *)wrapper)->nFailed;
  ((Wrapper *)wrapper)->cumPVal +=
      ((Wrapper *)wrapper)->nVal / (float)ccadical_vars(wrapper);
  ((Wrapper *)wrapper)->cumPFailed +=
      ((Wrapper *)wrapper)->nFailed / (float)ccadical_vars(wrapper);

  ((Wrapper *)wrapper)->nCalls++;
  int res = ((Wrapper *)wrapper)->solver->solve();
  ((Wrapper *)wrapper)->lastSat = (res == 10);
  ((Wrapper *)wrapper)->lastUnsat = (res == 20);
  ((Wrapper *)wrapper)->nSat += (res == 10);
  ((Wrapper *)wrapper)->nUnsat += (res == 20);
  ((Wrapper *)wrapper)->nVal = 0;
  ((Wrapper *)wrapper)->nFailed = 0;
  return res;
}

int ccadical_simplify (CCaDiCaL * wrapper) {
  return ((Wrapper*) wrapper)->solver->simplify ();
}

int ccadical_val (CCaDiCaL * wrapper, int lit) {
  ((Wrapper *)wrapper)->nVal++;
  return ((Wrapper*) wrapper)->solver->val (lit);
}

int ccadical_failed (CCaDiCaL * wrapper, int lit) {
  return ((Wrapper*) wrapper)->solver->failed (lit);
}

void ccadical_print_statistics (CCaDiCaL * wrapper) {
  const bool cModel = (((Wrapper *)wrapper)->lastSat) && (((Wrapper *)wrapper)->nVal);
  const bool cCore = (((Wrapper *)wrapper)->lastUnsat) && (((Wrapper *)wrapper)->nFailed);
  ((Wrapper *)wrapper)->readModel += cModel;
  ((Wrapper *)wrapper)->readCore += cCore;
  ((Wrapper *)wrapper)->cumVal += ((Wrapper *)wrapper)->nVal;
  ((Wrapper *)wrapper)->cumFailed += ((Wrapper *)wrapper)->nFailed;
  ((Wrapper *)wrapper)->cumPVal +=
      ((Wrapper *)wrapper)->nVal / (float)ccadical_vars(wrapper);
  ((Wrapper *)wrapper)->cumPFailed +=
      ((Wrapper *)wrapper)->nFailed / (float)ccadical_vars(wrapper);

  printf("nCalls: %d\n", ((Wrapper *)wrapper)->nCalls);
  printf(" nSat: %d\n", ((Wrapper *)wrapper)->nSat);
  printf(" readModel: %d\n", ((Wrapper *)wrapper)->readModel);
  printf(" nUnsat: %d\n", ((Wrapper *)wrapper)->nUnsat);
  printf(" readCore: %d\n", ((Wrapper *)wrapper)->readCore);
  printf("avgVal: %.2f\n", ((((Wrapper *)wrapper)->cumVal) /
                            ((float)((Wrapper *)wrapper)->readModel)));
  printf("avgPVal: %.2f\n", ((((Wrapper *)wrapper)->cumPVal) /
                             ((float)((Wrapper *)wrapper)->readModel)));
  printf("avgFailed: %.2f\n", ((((Wrapper *)wrapper)->cumFailed) /
                               ((float)((Wrapper *)wrapper)->readCore)));
  printf("avgPFailed: %.2f\n", ((((Wrapper *)wrapper)->cumPFailed) /
                                ((float)((Wrapper *)wrapper)->readCore)));

  ((Wrapper*) wrapper)->solver->statistics ();
}

void ccadical_terminate (CCaDiCaL * wrapper) {
  ((Wrapper*) wrapper)->solver->terminate ();
}

int64_t ccadical_active (CCaDiCaL * wrapper) {
  return ((Wrapper*) wrapper)->solver->active ();
}

int64_t ccadical_irredundant (CCaDiCaL * wrapper) {
  return ((Wrapper*) wrapper)->solver->irredundant ();
}

int ccadical_fixed (CCaDiCaL * wrapper, int lit) {
  return ((Wrapper*) wrapper)->solver->fixed (lit);
}

void ccadical_set_terminate (CCaDiCaL * ptr,
                             void * state, int (*terminate)(void *)) {
  Wrapper * wrapper = (Wrapper *) ptr;
  wrapper->terminator.state = state;
  wrapper->terminator.function = terminate;
  if (terminate) wrapper->solver->connect_terminator (wrapper);
  else wrapper->solver->disconnect_terminator ();
}

void ccadical_set_learn (CCaDiCaL * ptr,
                         void * state, int max_length,
			 void (*learn)(void * state, int * clause)) {
  Wrapper * wrapper = (Wrapper *) ptr;
  wrapper->learner.state = state;
  wrapper->learner.max_length = max_length;
  wrapper->learner.function = learn;
  if (learn) wrapper->solver->connect_learner (wrapper);
  else wrapper->solver->disconnect_learner ();
}


void ccadical_freeze (CCaDiCaL * ptr, int lit) {
  ((Wrapper*) ptr)->solver->freeze (lit);
}

void ccadical_melt (CCaDiCaL * ptr, int lit) {
  ((Wrapper*) ptr)->solver->melt (lit);
}

int ccadical_frozen (CCaDiCaL * ptr, int lit) {
  return ((Wrapper*) ptr)->solver->frozen (lit);
}

void ccadical_constrain (CCaDiCaL *wrapper, int lit){
  ((Wrapper*) wrapper)->solver->constrain (lit);
}

void ccadical_set_runid(CCaDiCaL *wrapper, int id) {
  ((Wrapper *) wrapper)->run_id = id;
}

long *ccadical_get_runtime_limit(CCaDiCaL *wrapper) {
  return &((Wrapper *) wrapper)->runtime_limit;
}

int ccadical_vars(CCaDiCaL *wrapper) {
  return ((Wrapper *) wrapper)->solver->vars ();
}

void ccadical_reserve(CCaDiCaL *wrapper, int min_max_var) {
  ((Wrapper *) wrapper)->solver->reserve (min_max_var);
}

void ccadical_clear_assumptions(CCaDiCaL *wrapper) {
  ((Wrapper *) wrapper)->assumptions.clear ();
}

void ccadical_update_core_assumptions(CCaDiCaL *wrapper) {
  auto &assumptions = ((Wrapper *) wrapper)->assumptions;
  auto i            = assumptions.begin (), j = i;
  for (; i != assumptions.end(); i++) {
    ((Wrapper *)wrapper)->nFailed++;
    if (!((Wrapper *)wrapper)->solver->failed(*i))
      continue;
    *j++ = *i;
  }
  assumptions.resize (j - assumptions.begin ());
}

int ccadical_get_core_assumptions(CCaDiCaL *wrapper, int **core_assumptions) {
  auto &assumptions = ((Wrapper *) wrapper)->assumptions;
  *core_assumptions = &assumptions[0];
  return assumptions.size ();
}

}
